#! /usr/bin/env python3

import curses
import argparse
import time
import collections
from itertools import cycle
import ai


def parse_minsec(s):
    p = s.split(":")
    return int(p[0]) * 60 + int(p[1])

def format_minsec(t):
    mins = t // 60
    secs = t % 60
    return "%02d:%02d" % (mins, secs)


parser = argparse.ArgumentParser(
    prog="jewels"
)

parser.add_argument(
    nargs="?",
    dest="fname",
    help="Input game file"
)

parser.add_argument(
    "-t",
    "--time",
    type=parse_minsec,
    dest="time",
    help="Time limit minutes:seconds",
)

class Point:

    def __init__(self, r, c):
        self.r = r
        self.c = c

    def add(self, dr, dc):
        return Point(self.r + dr, self.c + dc)


args = parser.parse_args()

def load_file(fname):
    lines = []
    with open(fname) as f:
        for line in f:
            lines.append([int(c) for c in line[:-1]])
    return lines[:8], [cycle(l) for l in lines[8:16]]

# Global state, yahoooooo!
field, nexts = load_file(args.fname)
nxats = [next(x) for x in nexts]
root = None
cursor = Point(0, 0)
picked = False
points = 0
endmsg = None
timelimitend = (time.time() + args.time) if args.time else None
hintmode = False
hintmos = None

def gemat(r, s, val):
    if val is None:
        # Not really supposed to happen ever
        root.addstr(r, s, "!!")
    else:
        root.addstr(r, s, "  ", curses.color_pair(val))

def draw():
    ofr = 0
    for i, nx in enumerate(nxats):
        gemat(ofr, i * 2, nx)

    ofr += 1

    root.addstr(ofr, 0, "-" * 16)

    ofr += 1
    for r, row in enumerate(field):
        for s, v in enumerate(row):
            gemat(r + ofr, s * 2, v)

    if hintmode:
        for hint in hintmos:
            if hint[2] == ai.DOWN:
                root.addstr(ofr + hint[0], hint[1] * 2 + 0, "⇓",
                            curses.color_pair(field[hint[0]][hint[1]]))
            elif hint[2] == ai.RIGHT:
                root.addstr(ofr + hint[0], hint[1] * 2 + 1, "⇒",
                            curses.color_pair(field[hint[0]][hint[1]]))

    root.addstr(cursor.r + ofr, cursor.c * 2, "==" if picked else "<>",
                curses.color_pair(field[cursor.r][cursor.c]))

    ofr += 8
    root.addstr(ofr, 0, "Points: % 4d" % points)
    ofr += 1
    root.addstr(ofr, 0, format_minsec(timelimitend - time.time()) if timelimitend else "Unlimited time")

def fillup():
    # Fill with next stones as motion_points does not do that
    for s in range(len(field[0])): # _columns_ first here!
        for r in range(len(field) - 1, -1, -1): # From bottom upwards
            if field[r][s] is None:
                # Fill with next stone
                field[r][s] = nxats[s]
                nxats[s] = next(nexts[s])

def try_motion(dr):
    global field, points, picked, endmsg, hintmos
    try:
        pts, nf = ai.motion_points(field, cursor.r, cursor.c, dr)
    except IndexError:
        return
    if pts == 0:
        return
    # Valid motion, perform it
    field = nf
    points += pts
    picked = False
    # Essentially, we need to continue filling until there are no connections
    while pts:
        fillup()
        pts = ai.remove_conns(field)
        points += pts

    # Next, we need to check whether there are any more available motions and
    # end the game if not
    hintmos = list(ai.available_motions(field))
    if not hintmos:
        endmsg = "No more valid moves available!"
        raise KeyboardInterrupt

hintmos = list(ai.available_motions(field))
try:
    root = curses.initscr()
    curses.curs_set(0)
    curses.noecho()
    root.timeout(300)

    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_RED)
    curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_GREEN)
    curses.init_pair(3, curses.COLOR_WHITE, curses.COLOR_BLUE)
    curses.init_pair(4, curses.COLOR_BLACK, curses.COLOR_WHITE)
    curses.init_pair(5, curses.COLOR_WHITE, curses.COLOR_YELLOW)
    curses.init_pair(6, curses.COLOR_WHITE, curses.COLOR_MAGENTA)

    while True:
        draw()
        root.refresh()
        inp = root.getch()
        if inp in {ord("W"), ord("w")}:
            # Move up
            if picked:
                try_motion(ai.UP)
            else:
                if cursor.r > 0:
                    cursor = cursor.add(-1, 0)
        elif inp in {ord("S"), ord("s")}:
            # Move down
            if picked:
                try_motion(ai.DOWN)
            else:
                if cursor.r < len(field) - 1:
                    cursor = cursor.add(1, 0)
        elif inp in {ord("A"), ord("a")}:
            # Move left
            if picked:
                try_motion(ai.LEFT)
            else:
                if cursor.c > 0:
                    cursor = cursor.add(0, -1)
        elif inp in {ord("D"), ord("d")}:
            # Move right
            if picked:
                try_motion(ai.RIGHT)
            else:
                if cursor.c < len(field[0]) - 1:
                    cursor = cursor.add(0, 1)
        elif inp == ord(" "):
            # Pickup/dropped jewel
            picked = not picked
        elif inp in {ord("Q"), ord("q")}:
            raise KeyboardInterrupt
        elif inp == ord("?"):
            hintmode = not hintmode
        else:
            # Timeout, check timer
            if timelimitend is not None and time.time() > timelimitend:
                endmsg = "Time limit exceeded!"
                raise KeyboardInterrupt
except KeyboardInterrupt:
    pass
finally:
    curses.endwin()
    if endmsg:
        print(endmsg)
    print("Points: %d" % points)
